#include <stdio.h>
#include <math.h>

int Interrupt();

double min(double,double), max(double, double);
int isign(double);

void fcn(double*, double*), jac(double*, double*);
double* dcalloc(int, char*);
int* icalloc(int, char*);

int solve(int n, double *a, int *ip, double *b);


/* ********************************************** */
/* *   copyright: b.a. gottwald and g. wanner   * */
/* *     a reliable rosenbrock integrator       * */
/* *     for stiff differential equations       * */
/* *          computing 26: 355 (1981)          * */
/* ********************************************** */



int decompose(int n, double *a, int *ip, int *ier) {
  double t;
  int i, j, k, kp1, m, nm1;

  if (Interrupt())
    return 0;

  *ier = 0;
  ip[n-1] = 1; /*  let op : was 'n' */
  if (n > 1) {
    nm1 = n - 1;
    for (k = 1; k <= nm1; ++k) {
      kp1 = k + 1;
      m = k;
      for (i = kp1; i <= n; ++i) {
	if(fabs(a[(i-1)*n + k-1]) > fabs(a[(m-1)*n + k-1]))
	  m = i;
      }
      ip[k-1] = m;
      t = a[(m-1)*n + k-1];
      if (m != k) {
	ip[n-1] = -ip[n-1];
	a[(m-1)*n + k-1] = a[(k-1)*n + k-1];
	a[(k-1)*n + k-1] = t;
      }
      if (isign(t) != 0) {
	t = 1 / t;
      } else {
	*ier = k;
	ip[n-1] = 0;
	return 1;
      }
      for (i = kp1; i <= n; ++i)
	a[(i-1)*n + k-1] *= -t;
      for (j = kp1; j <= n; ++j) {
	t = a[(m-1)*n + j-1];
	a[(m-1)*n + j-1] = a[(k-1)*n + j-1];
	a[(k-1)*n + j-1] = t;
	if (t != 0.)
	  for (i = kp1; i <= n; ++i)
	    a[(i-1)*n + j-1] += a[(i-1)*n + k-1] * t;
      }
    }
  }
  if (isign(a[(n-1)*n + n-1]) == 0) {
    *ier = n;
    ip[n] = 0;
  }
  
  return 1;
} /* decompose */

int row4a(int n, double *t, double *y, double tend, double tol, double hmax, double h, int *pas, int *rep) {
    /* Initialized data */

    static double a32 = .0730795420615381;
    static double c21 = -1.94347441894707;
    static double c31 = .416957530989189;
    static double c32 = 1.32396782072923;
    static double c41 = 1.51951325778448;
    static double c42 = 1.35370815030093;
    static double c43 = -.854151495257539;
    static double b1 = .729044879960308;
    static double b2 = .0541069773272405;
    static double b3 = .281599362440017;
    static double b4 = .25;
    static double e4 = .25;
    static double e1 = -.0190858871999474;
    static double e2 = .255608791716455;
    static double e3 = -.0863816280897592;
    static double a21 = .438;
    static double a31 = .938948678483428;

    double uround, told=0., tnew, hold=0., hnew, s, sk, est, fac, facmin, facmax;
    int i, j, ifail, final=0;

    static int *pivot = NULL;
    static double *yold = NULL, *ynew = NULL, *dy = NULL, *dy1 = NULL, *k1 = NULL, *k2 = NULL, *k3 = NULL, *k4 = NULL, *ddy = NULL,  *e = NULL;

    if (Interrupt())
      return 0;

    if(pivot == NULL)  pivot = icalloc(n, "row4a");
    if(yold == NULL)  yold = dcalloc(n, "row4a");
    if(ynew == NULL)  ynew = dcalloc(n, "row4a");
    if(dy == NULL)  dy = dcalloc(n, "row4a");
    if(dy1 == NULL)  dy1 = dcalloc(n, "row4a");
    if(k1 == NULL)  k1 = dcalloc(n, "row4a");
    if(k2 == NULL)  k2 = dcalloc(n, "row4a");
    if(k3 == NULL)  k3 = dcalloc(n, "row4a");
    if(k4 == NULL)  k4 = dcalloc(n, "row4a");
    if(ddy == NULL)  ddy = dcalloc(n*n, "row4a");
    if(e == NULL)  e = dcalloc(n*n, "row4a");


    uround = max(1.73e-18, tol * 1e-8);
    facmin = .125;
    facmax = 2.;
    h = fabs(h);
    *pas = *rep = 0;

 dacapo:
    fcn(y, dy1);
    jac(y, ddy);
 denovo:
    for (i = 1; i <= n; ++i) {
	for (j = 1; j <= n; ++j)
	    e[(i-1)*n + j-1] = -h*ddy[(i-1)*n + j-1]*.395;
	e[(i-1)*n + i-1] += 1;
    }
    if (!decompose(n, e, pivot, &ifail))
      return 0;
    if (ifail != 0) {
	h *= .5;
	goto denovo;
    }
    for (i = 1; i <= n; ++i)
      k1[i-1] = h*dy1[i-1];
    solve(n, e, pivot, k1);

    for (i = 1; i <= n; ++i)
	ynew[i-1] = y[i-1] + a21 * k1[i-1];
    fcn(ynew, dy);

    for (i = 1; i <= n; ++i)
	k2[i-1] = h*dy[i-1] + c21*k1[i-1];
    solve(n, e, pivot, k2);

    for (i = 1; i <= n; ++i)
	ynew[i-1] = y[i-1] + a31*k1[i-1] + a32*k2[i-1];
    fcn(ynew, dy);

    for (i = 1; i <= n; ++i)
	k3[i-1] = h * dy[i-1] + c31*k1[i-1] + c32*k2[i-1];
    solve(n, e, pivot, k3);

    for (i = 1; i <= n; ++i)
	k4[i-1] = h*dy[i-1] + c41*k1[i-1] + c42*k2[i-1] + c43*k3[i-1];
    solve(n, e, pivot, k4);

    for (i = 1; i <= n; ++i)
	ynew[i-1] = b1*k1[i-1] + b2*k2[i-1] + b3*k3[i-1] + b4*k4[i-1] + y[i-1];

    if (Interrupt())
      return 0;
    if (final) {
	*t = tend;
	h = hnew;
	for (i = 1; i <= n; ++i)
	    y[i-1] = ynew[i-1];
	return 1;
    }
    tnew = *t+h;
    est = uround;
    for (i = 1; i <= n; ++i) {
	s = e1*k1[i-1] + e2*k2[i-1] + e3*k3[i-1] + e4*k4[i-1];
	sk = max((double)1,max(fabs(y[i-1]), fabs(ynew[i-1])));
	est = max(est, fabs(s)/sk);
    }
    fac = 0.9*pow(tol/est, (double)0.25) ;

    hnew = min(hmax, h*min(facmax, max(facmin, fac)));
    if (est > tol) {
      if (*pas > 0)
	(*rep)++;
      h = hnew;
      if (hnew >= hold) {
	facmax = 1.;
	goto denovo;
      } 
      else {
	(*pas)--;
	(*rep)++;
	facmax = .95;
	hold = 0.;
	*t = told;
	for (i = 1; i <= n; ++i)
	  y[i-1] = yold[i-1];
	goto dacapo;
      }
    }
    if (*t > tend) {
      final = 1;
      h = tend-told;
      hold = 0.;
      *t = told;
      for (i = 1; i <= n; ++i)
	y[i-1] = yold[i-1];
      goto dacapo;
    }
    if (isign(facmax - .95) == 0)
	facmax = 0.9;
    else
	facmax = 1.5;
    facmin = .5;
    (*pas)++;
    told = *t;
    *t = tnew;
    hold = h;
    h = hnew;
    for (i = 1; i <= n; ++i) {
      yold[i-1] = y[i-1];
      y[i-1] = ynew[i-1];
    }
    goto dacapo;
}

int solve(int n, double *a, int *ip, double *b) {
  int i, k, kb, km1, m, nm1;
  double t;

  if (Interrupt())
    return 0;

  if (n > 1) {
    nm1 = n-1;
    for (k = 1; k <= nm1; ++k) {
      m = ip[k-1];
      t = b[m-1];
      b[m-1] = b[k-1];
      b[k-1] = t;
      for (i=k+1; i <= n; ++i)
	b[i-1] += a[(i-1)*n + k-1]*t;
    }
    for (kb = 1; kb <= nm1; ++kb) {
      km1 = n - kb;
      k = km1 + 1;
      b[k-1] /= a[(k-1)*n + k-1];
      t = -b[k-1];
      for (i = 1; i <= km1; ++i)
	b[i-1] += a[(i-1)*n + k-1]*t;
    }
  }
  b[0] /= a[0];
  return 1;
}
